#ifndef SATTYPES_HPP_
#define SATTYPES_HPP_
#define types_hpp_version "$Revision: 1.4 $"

#include <vector>
#include "Assertion.hpp"
#include <stdio.h>
#include <stdint.h>
#include "OFStream.hpp"
#include "LazyFraction.hpp"
#include "Heap.hpp"
#include "EInfo.hpp"


using std::vector;


class Literal;
class SimplyFire;





/**Class to represent boolean values and a third state 'Unknown'*/
class TrBool{
    /**The internal value*/
    uint8_t val;

//public members
public:
    /**Constructs an object with unknown value*/
    inline              TrBool();

    /**Constructs an object with the specified value.
     * @param i: a value from {0,1,2}*/
    inline explicit     TrBool(uint8_t i);

    /**Copy construction*/
    inline              TrBool(const TrBool& b);

    inline uint8_t      toInt()                     const;
    inline string		toStrng()					const;
    inline bool         isTrue()                    const;
    inline bool         isFalse()                   const;
    inline bool         isUnknown()                 const;
    inline bool         isKnown()                   const;
    inline bool         operator==(const TrBool& o) const;
    inline bool         operator!=(const TrBool& o) const;
};


/**Class to represent variables. All additional information about a variable are stored outside: in an instance of
 * ConstraintDB*/
class Variable{

	uint32_t ID;

public:

    /**Construct a variable from the given id*/
    inline explicit     Variable(const uint32_t i);
    inline explicit		Variable(){};

    inline uint32_t     idx()							const;
    inline Literal      toLit(const bool pos)			const;
    inline Literal      toPosLit()						const;
    inline Literal      toNegLit()						const;
    inline uint         toDIMACS()						const;
    inline bool         operator<(const Variable& v)	const;
    inline bool         operator>=(const Variable& v)	const;
    inline bool         operator!=(const Variable& v)	const;
    inline bool         operator==(const Variable& v)	const;

    friend class Literal;
};


/**Class to represent literals. All aditional information are stored elsewhere: in an instance of class ConstraintDB*/
class Literal{

	uint32_t ID;

public:
    /**Return a literal object from the given DIMACS number*/
    inline static Literal DIMACS2Literal(int32_t d);
    inline static bool nrmlLit(const Literal& a);

    /**Construct an invalid literal*/
    //inline                  Literal();

    inline explicit         Literal(const uint32_t i);
    inline explicit			Literal(){};

    inline uint32_t         idx()			const;
    inline Literal          flipped()		const;
    inline Variable         toVar()			const;
    inline bool             isPos()			const;
    inline bool             isNeg()			const;
    inline TrBool           pol()			const;
    inline TrBool           flpdPolarity()       const;
    inline int              toDIMACS()              const;
    inline bool             operator<(const Literal& l) const;
    inline bool             operator>(const Literal& l) const;
    inline bool             operator!=(const Literal& l) const;
    inline bool             operator==(const Literal& l)const;

    friend class Variable;
    friend class SimplyFire;
};

typedef LazyFraction<7,uint16_t,uint16_t> var_act_t;
struct VarActComp{
	//Variable ext_border;
	bool better(const Variable& va, const var_act_t& a,const Variable& vb, const var_act_t& b)const{
		return a > b;}
		//const bool anrml = va < ext_border;
		//const bool bnrml = vb < ext_border;
		//return (anrml == bnrml) ?  a > b : bnrml;}

	//VarActComp(Variable border):ext_border(border){}
	VarActComp(){}
};
typedef BumpHeap< Variable, var_act_t,VarActComp > var_hp_t;




//Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
const Literal   INVALID_LIT             (~zero32);
const Variable  INVALID_VAR             (~zero32);
const Variable	FIX_VAR					(0);
const Literal	TRUE_LIT				= FIX_VAR.toPosLit();
const Literal	FALSE_LIT				= FIX_VAR.toNegLit();


const uint8_t   FALSE_TRBOOL_VAL      	= 0;
const uint8_t   TRUE_TRBOOL_VAL       	= 1;
const uint8_t   TRBOOL_FLIP_MASK      	= 1;
const uint8_t   UNKNOWN_TRBOOL_VAL    	= 2;
const TrBool    TRUE_TrBool(TRUE_TRBOOL_VAL);
const TrBool    FALSE_TrBool(FALSE_TRBOOL_VAL);
const TrBool    UNKNOWN_TrBool(UNKNOWN_TRBOOL_VAL);



//TBool Implementation
            TrBool::TrBool()                            :val(UNKNOWN_TRBOOL_VAL){ }
            TrBool::TrBool(const TrBool& b)             :val(b.val)             { }
            TrBool::TrBool(uint8_t i)                   :val(i)     {
                            asrt(i == TRUE_TRBOOL_VAL || i == FALSE_TRBOOL_VAL || i == UNKNOWN_TRBOOL_VAL);}

uint8_t     TrBool::toInt()                 const       {return val;}
bool        TrBool::isTrue()                const       {return val == TRUE_TRBOOL_VAL;}
bool        TrBool::isFalse()               const       {return val == FALSE_TRBOOL_VAL;}
bool        TrBool::isUnknown()             const       {return val == UNKNOWN_TRBOOL_VAL;}
bool        TrBool::isKnown()               const       {return val != UNKNOWN_TRBOOL_VAL;}
bool        TrBool::operator==(const TrBool& o) const   {return val == o.val;}
bool        TrBool::operator!=(const TrBool& o) const   {return val != o.val;}
string		TrBool::toStrng()				const		{return isUnknown() ? "unknown" : (isTrue()?"true":"false");}



//Variable Implementation
            Variable::Variable(const uint32_t i) :ID(i)          { }
Literal     Variable::toLit(const bool pos)			const	{return Literal((ID << 1) + pos);}
Literal     Variable::toPosLit()                    const   {return Literal((ID << 1) + 1);}
Literal     Variable::toNegLit()                    const   {return Literal(ID << 1);}
uint32_t    Variable::idx()                         const   {return ID;}
uint32_t    Variable::toDIMACS()                    const   {return ID;}//ID+1;}
bool        Variable::operator<(const Variable& v)  const   {return ID < v.ID;}
bool        Variable::operator>=(const Variable& v) const   {return ID >= v.ID;}
bool        Variable::operator!=(const Variable& v) const   {return ID != v.ID;}
bool        Variable::operator==(const Variable& v) const   {return ID == v.ID;}

//Literal Implementation
            //Literal::Literal()              		:ID(~zero32){ }
            Literal::Literal(const uint32_t i)   	:ID(i)      { }
uint32_t    Literal::idx()                          const   {return ID;}
Literal     Literal::flipped()                      const   {return Literal(ID ^ 1);}
Variable    Literal::toVar()                        const   {return Variable(ID >> 1);}
bool        Literal::isPos()                        const   {return ID & 1;}
bool        Literal::isNeg()                        const   {return !(ID & 1);}
int         Literal::toDIMACS()                     const   {return isPos() ? (ID>>1) : -(ID>>1) ;}
TrBool      Literal::pol()                          const   {return TrBool(ID & 1);}
TrBool      Literal::flpdPolarity()                 const   {return TrBool(!(ID&1));}
bool        Literal::operator==(const Literal& l)   const   {return ID == l.ID; }
bool        Literal::operator<(const Literal& l)    const   {return ID < l.ID;}
bool        Literal::operator>(const Literal& l)    const   {return ID > l.ID;}
bool        Literal::operator!=(const Literal& l)   const   {return ID != l.ID;}

bool 		Literal::nrmlLit(const Literal& a)		{
			asrt(TRUE_LIT.ID>FALSE_LIT.ID); return a.ID > TRUE_LIT.ID;}

Literal     Literal::DIMACS2Literal(int32_t d){
    asrt(d != 0);
    //return Literal(d < 0 ? (-d-1)*2 : d*2-1);
    uint32_t m(d>>31);
    return Literal(((d^m)<<1)-m+1);
}






/*

 uint32_t Cls::computeKey(const Literal* lits, uint32_t sz){ asrt(sz>2);
	uint32_t res = lits[--sz].idx();
	while(--sz){res ^= lits[sz].idx();} //for all but the first one
	return res;
}

Literal	Cls::recompOneWatch(const uint32_t key){
	uint32_t k = watches;
	for(uint32_t t=innerSize();t--;){
		k ^= literals[t].idx();}
	return Literal(k ^ key);
}


void	Cls::getAllLits(const uint32_t key,vector<Literal>& lits){
	uint32_t k = watches;
	for(uint32_t t=innerSize();t--;){
		k 		^= literals[t].idx();
		lits	.push_back(literals[t]);
	}
	Literal w0 	= Literal(k ^ key);
	lits		.push_back(w0);
	lits		.push_back(otherWatcher(w0));
}*/



/*
void			Cls::removeAt(uint32_t idx){
	asrt(idx > 1  && idx < size());
	literals[idx] = literals[size() - 1];
	--siz;
	clearStamp();
}


void			Cls::shrinkToSize(uint32_t new_sz){
	asrt(new_sz > 2 && new_sz < size());

	size29_watched_flag_learnt = (new_sz<< 3) | (size29_watched_flag_learnt & 7) ;
	clearStamp();
	asrt(size() == new_sz);
}*/



/*

bool Cls::asrtSorted() const{
	Literal last = INVALID_LIT;
	for(uint32_t i = size();i--;){
		if(literals[i].idx() >= last.idx()) return false;
		last = literals[i];
	}
	return true;
}


void Cls::print(OFStream& out){
    for(uint32_t l = 0; l != size(); ++l){
        out.writeString("%i ",literals[l].toDIMACS() );
    }
}
*/






#endif /*SATTYPES_HPP_*/
